import torch
import random
from torchvision.models import resnet18, resnet50, mobilenet_v2, densenet169, densenet121, densenet201, googlenet
import argparse
import os
import numpy as np
from define_data import get_data
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F

class LinearClassifier(nn.Module):
    """Linear classifier"""
    def __init__(self, feat_dim, num_classes):
        super(LinearClassifier, self).__init__()
        self.linear = nn.Linear(feat_dim, num_classes, bias = True)

    def forward(self, features):
            return self.linear(features)
    
def train_epoch(args, FModel, TModel, dataloader, optimizer, criterion, cur_epoch):

    TModel.train()
    # One epoch
    top1 = AverageMeter()
    top5 = AverageMeter()
    running_loss = 0.0
    for batch_idx, (inputs, targets) in enumerate(dataloader):

        inputs, targets = inputs.to(args.device), targets.to(args.device)
        
        feature = FModel(inputs)
        if args.model == "inception_v3":
            outputs = TModel(feature) #outputs = TModel(feature[0])
        else:
            outputs = TModel(feature)
        
        loss = criterion(outputs, targets)
        
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        
        prec1, prec5 = compute_accuracy(outputs.data, targets.data, topk=(1, 5))
        top1.update(prec1.item(), inputs.size(0))
        top5.update(prec5.item(), inputs.size(0))

        # print statistics
        running_loss += loss.item()
        if batch_idx % 100 == 0:    # print every 100 mini-batches
            print(f'[Epoch {cur_epoch}, {batch_idx + 1:5d}] loss: {running_loss / 100 / args.batch_size:.3f}, acc: {top1.avg}')
            running_loss = 0.0
        
    return top1.avg, top5.avg

def evaluate_epoch(args, FModel, TModel, dataloader, criterion):

    TModel.eval()
    # One epoch
    test_top1 = AverageMeter()
    test_top5 = AverageMeter()
    running_loss = 0.0
    with torch.no_grad():
        for batch_idx, (inputs, targets) in enumerate(dataloader):

            inputs, targets = inputs.to(args.device), targets.to(args.device)
            
            with torch.no_grad():
                feature = FModel(inputs)
                if args.model == "inception_v3":
                    outputs = TModel(feature) #outputs = TModel(feature[0])
                else:
                    outputs = TModel(feature)

            loss = criterion(outputs, targets)

            prec1, prec5 = compute_accuracy(outputs.data, targets.data, topk=(1, 5))
            test_top1.update(prec1.item(), inputs.size(0))
            test_top5.update(prec5.item(), inputs.size(0))

    return test_top1.avg, test_top5.avg
    
def compute_accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].reshape(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res

class AverageMeter(object):
    """Computes and stores the average and current value
       Imported from https://github.com/pytorch/examples/blob/master/imagenet/main.py#L247-L262
    """
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count
        
def set_seed(manualSeed=666):
    random.seed(manualSeed)
    np.random.seed(manualSeed)
    torch.manual_seed(manualSeed)
    torch.cuda.manual_seed(manualSeed)
    torch.cuda.manual_seed_all(manualSeed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    os.environ['PYTHONHASHSEED'] = str(manualSeed)
    

def parse_eval_args():
    parser = argparse.ArgumentParser()

    # parameters
    # Model Selection
    parser.add_argument('--model', type=str, default='resnet18')
    parser.add_argument('--feat_dim', type=int, default=512) # 512 for ResNet18, 2048 for ResNet50

    # Hardware Setting
    parser.add_argument('--gpu_id', type=int, default=0)
    parser.add_argument('--seed', type=int, default=6)

    # Directory Setting
    # Note dataset mean which dataset to transfer learning on
    parser.add_argument('--dataset', type=str, choices=['cifar10','cifar100','pet','dtd','aircraft'], default='cifar10')
    parser.add_argument('--data_dir', type=str, default='<path to folder where data should be saved>')
    parser.add_argument('--save_path', type=str, default='saved/')

    # Learning Options
    parser.add_argument('--epochs', type=int, default=200, help='Max Epochs')
    parser.add_argument('--batch_size', type=int, default=128, help='Batch size')

    args = parser.parse_args()

    return args

def main():
    args = parse_eval_args()

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    args.device = device

    set_seed(args.seed)
    
    # Dataset part
    print(f"Using dataset {args.dataset}")
    print()
    trainloader, testloader, num_classes = get_data(args.dataset, args.data_dir,
                                                    args.batch_size, do_transform = True)
    
    if args.model == "resnet18":
        model_train = resnet18(weights='IMAGENET1K_V1').to(device)
        model_train.fc = nn.Sequential()
    elif args.model == "resnet50":
        model_train = resnet50(weights='IMAGENET1K_V1').to(device)
        model_train.fc = nn.Sequential()
    elif args.model == "densenet121": # 1024
        model_train = densenet121(weights='IMAGENET1K_V1').to(device)
        model_train.classifier = nn.Sequential()
    elif args.model == "densenet169": # 1664
        model_train = densenet169(weights='IMAGENET1K_V1').to(device)
        model_train.classifier = nn.Sequential()
    elif args.model == "densenet201": # 1920
        model_train = densenet201(weights='IMAGENET1K_V1').to(device)
        model_train.classifier = nn.Sequential()
    elif args.model == "inception_v1": #1024
        model_train = googlenet(weights='IMAGENET1K_V1').to(device)
        model_train.fc = nn.Sequential()
    elif args.model == "mobilenet": #1280
        model_train = mobilenet_v2(weights='IMAGENET1K_V1').to(device)
        model_train.classifier = nn.Sequential()
    else:
        raise ValueError(f"Model type {args.model} not supported")
    
    model_ft = LinearClassifier(feat_dim=args.feat_dim, num_classes=num_classes).to(device)
    
    model_train.eval()
    model_train.requires_grad = False
    
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.SGD(model_ft.parameters(), lr=0.1, momentum = 0.9, weight_decay=1e-4)
    scheduler = optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=args.epochs, eta_min=0.0001)
    
    epochs = args.epochs
    best_test_acc = 0.0
    train_accs = []
    test_accs = []
    is_best = False 
    checkpoint_dir = args.save_path + f"{args.model}_{args.dataset}/"
    if not os.path.exists(checkpoint_dir):
        os.makedirs(checkpoint_dir, exist_ok=True)
        
    for epoch in range(epochs):
        train_top1, train_top5 = train_epoch(args, model_train, model_ft, trainloader, optimizer, criterion, epoch)
        test_top1, test_top5 = evaluate_epoch(args, model_train, model_ft, testloader, criterion)
        train_accs.append(train_top1)
        test_accs.append(test_top1)
        scheduler.step()

        if test_top1 > best_test_acc:
            is_best = True
            best_test_acc = test_top1
        else:
            is_best = False
        print(f"Finish Epoch {epoch}, Training Acc: {train_top1}, Test Acc: {test_top1}, Current best Test Acc: {best_test_acc}, current LR: {scheduler.get_lr()}")
        
        state = {
                    'arch': args.model,
                    'epoch': epoch,
                    'state_dict': model_train.state_dict(),
                    'state_dict_ft': model_ft.state_dict(),
                    'optimizer': optimizer.state_dict(),
                    'test_acc': test_top1
                }
        if is_best:
            print("Save current model (best)")
            path = checkpoint_dir + 'model_best.pth'
            torch.save(state, path)
        elif (epoch+1) % 50 == 0:
            print("Save current model (epoch)")
            path = checkpoint_dir + f'model_epoch_{epoch}.pth'
            torch.save(state, path)
    
    print(f"Training Accs are: {train_accs}")
    print(f"Test Accs are: {test_accs}")

if __name__ == "__main__":
    main()