import pickle
import torch
import torch.nn as nn
import torch.nn.functional as F
import scipy.linalg as scilin
from torchvision.models import resnet18, resnet50, mobilenet_v2, densenet169, densenet121, densenet201, googlenet
import argparse
import numpy as np
import matplotlib.pyplot as plt
from define_data import get_data
from utils import CDNV

class LinearClassifier(nn.Module):
    """Linear classifier"""
    def __init__(self, feat_dim, num_classes):
        super(LinearClassifier, self).__init__()
        self.linear = nn.Linear(feat_dim, num_classes, bias = True)

    def forward(self, features):
            return self.linear(features)

def compute_accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].reshape(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res

class AverageMeter(object):
    """Computes and stores the average and current value
       Imported from https://github.com/pytorch/examples/blob/master/imagenet/main.py#L247-L262
    """
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

def parse_eval_args():
    parser = argparse.ArgumentParser()

    # parameters
    # Model Selection
    parser.add_argument('--model', type=str, default='resnet18') # Model type
    parser.add_argument('--after_ft', dest='after_ft', action='store_true') # Validate model after fine tune
    parser.add_argument('--feat_dim', type=int, default=512) # 512 for ResNet18, 2048 for ResNet50

    # Hardware Setting
    parser.add_argument('--gpu_id', type=int, default=0)

    # Directory Setting
    parser.add_argument('--dataset', type=str, choices=['mnist', 'cifar10', 'cifar10_random', 'miniimagenet','cifar100','pet','dtd','aircraft'], default='cifar10')
    parser.add_argument('--data_dir', type=str, default='<path to folder where data should be saved>')
    parser.add_argument('--save_path', type=str, default="saved/")
    parser.add_argument('--load_path', type=str, default=None)

    # Learning Options
    parser.add_argument('--batch_size', type=int, default=128, help='Batch size')

    args = parser.parse_args()

    return args

def split_array(input_array, batchsize=128):
    input_size = input_array.shape[0]
    num_splits, res_splits = input_size // batchsize, input_size % batchsize
    output_array_list = list()
    if res_splits == 0:
        output_array_list = np.split(input_array, batchsize, axis=0)
    else:
        for i in range(num_splits):
            output_array_list.append(input_array[i*batchsize:(i+1)*batchsize])

        output_array_list.append(input_array[num_splits*batchsize:])

    return output_array_list

def compute_info(args, model_train, dataloader):
    model_train
    num_data = 0
    mu_G = 0
    mu_c_dict = dict()
    num_class_dict = dict()
    before_class_dict = dict()
    for batch_idx, (inputs, targets) in enumerate(dataloader):

        inputs, targets = inputs.to(args.device), targets.to(args.device)

        with torch.no_grad():
            features = model_train(inputs)
        
        # Need to normalize feature
        #features = F.normalize(features, dim=1)
        # Need to normalize feature
        
        if batch_idx == 0:
            print(f"Feature shape: {features.shape}")
        mu_G += torch.sum(features, dim=0)

        for b in range(len(targets)):
            y = targets[b].item()
            if y not in mu_c_dict:
                mu_c_dict[y] = features[b, :]
                before_class_dict[y] = [features[b, :].detach().cpu().numpy()]
                num_class_dict[y] = 1
            else:
                mu_c_dict[y] += features[b, :]
                before_class_dict[y].append(features[b, :].detach().cpu().numpy())
                num_class_dict[y] = num_class_dict[y] + 1

        num_data += targets.shape[0]

    mu_G /= num_data
    for i in range(len(mu_c_dict.keys())):
        mu_c_dict[i] /= num_class_dict[i]

    return mu_G, mu_c_dict, before_class_dict

def compute_Sigma_W(args, before_class_dict, mu_c_dict, batchsize=128):
    num_data = 0
    Sigma_W = 0

    for target in before_class_dict.keys():
        class_feature_list = split_array(np.array(before_class_dict[target]), batchsize=batchsize)
        for features in class_feature_list:
            features = torch.from_numpy(features).to(args.device)
            Sigma_W_batch = (features - mu_c_dict[target].unsqueeze(0)).unsqueeze(2) @ (
                        features - mu_c_dict[target].unsqueeze(0)).unsqueeze(1)
            Sigma_W += torch.sum(Sigma_W_batch, dim=0)
            num_data += features.shape[0]

    Sigma_W /= num_data
    return Sigma_W.detach().cpu().numpy()

def compute_Sigma_B(mu_c_dict, mu_G):
    Sigma_B = 0
    K = len(mu_c_dict)
    for i in range(K):
        Sigma_B += (mu_c_dict[i] - mu_G).unsqueeze(1) @ (mu_c_dict[i] - mu_G).unsqueeze(0)

    Sigma_B /= K

    return Sigma_B.cpu().numpy()

def main():
    args = parse_eval_args()

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    args.device = device

    # Dataset part
    print(f"Using dataset {args.dataset}")
    print()
    trainloader, testloader, num_classes = get_data(args.dataset, args.data_dir, 
                                                    args.batch_size, do_transform = False)
    
    print()
    print(f"We are validating using model type {args.model}!")
    print()
    
    if args.model == "resnet18":
        model_train = resnet18(weights='IMAGENET1K_V1').to(device)
        model_train.fc = nn.Sequential()
    elif args.model == "resnet50":
        model_train = resnet50(weights='IMAGENET1K_V1').to(device)
        model_train.fc = nn.Sequential()
    elif args.model == "densenet121": # 1024
        model_train = densenet121(weights='IMAGENET1K_V1').to(device)
        model_train.classifier = nn.Sequential()
    elif args.model == "densenet169": # 1664
        model_train = densenet169(weights='IMAGENET1K_V1').to(device)
        model_train.classifier = nn.Sequential()
    elif args.model == "densenet201": # 1920
        model_train = densenet201(weights='IMAGENET1K_V1').to(device)
        model_train.classifier = nn.Sequential()
    elif args.model == "inception_v1": #1024
        model_train = googlenet(weights='IMAGENET1K_V1').to(device)
        model_train.fc = nn.Sequential()
    elif args.model == "mobilenet":
        model_train = mobilenet_v2(weights='IMAGENET1K_V1').to(device)
        model_train.classifier = nn.Sequential()
    else:
        raise ValueError(f"Model type {args.model} not supported")
        
    model_ft = LinearClassifier(feat_dim=args.feat_dim, num_classes=num_classes).to(device)

    info_dict = {
                 'collapse_metric': [],
                 'collapse_metric_test': [],
                 'cdnv': [],
                 'cdnv_test': [],
                 }
    
    if args.after_ft:
        print(f"Load model from {args.load_path}/model_best.pth")
        #checkpoint = torch.load(args.load_path + 'model_best' + '.pth', map_location=device)
        checkpoint = torch.load(args.load_path + 'model_epoch_199' + '.pth', map_location=device)
        model_train.load_state_dict(checkpoint["state_dict"])
        model_ft.load_state_dict(checkpoint["state_dict_ft"])
    
    model_train.eval()
    model_ft.eval()

    mu_G_train, mu_c_dict_train, before_class_dict_train = compute_info(args, model_train, trainloader)
    mu_G_test, mu_c_dict_test, before_class_dict_test = compute_info(args, model_train, testloader)
    
    cdnv = CDNV(before_class_dict_train)
    info_dict['cdnv'].append(cdnv)
    cdnv_test = CDNV(before_class_dict_test)
    info_dict['cdnv_test'].append(cdnv_test)

    Sigma_W = compute_Sigma_W(args, before_class_dict_train, mu_c_dict_train, batchsize=args.batch_size)
    # Sigma_W_test_norm = compute_Sigma_W(args, model, fc_features, mu_c_dict_train, testloader, isTrain=False)
    Sigma_W_test = compute_Sigma_W(args, before_class_dict_test, mu_c_dict_test, batchsize=args.batch_size)
    Sigma_B = compute_Sigma_B(mu_c_dict_train, mu_G_train)
    Sigma_B_test = compute_Sigma_B(mu_c_dict_test, mu_G_test)

    collapse_metric = np.trace(Sigma_W @ scilin.pinv(Sigma_B)) / len(mu_c_dict_train)
    collapse_metric_test = np.trace(Sigma_W_test @ scilin.pinv(Sigma_B_test)) / len(mu_c_dict_test)
    print("NC1", collapse_metric)
    print("NC1_test", collapse_metric_test)
    
    info_dict['collapse_metric'].append(collapse_metric)
    info_dict['collapse_metric_test'].append(collapse_metric_test)
    
    if args.after_ft:
        save_path = args.load_path + "/"
    else:
        save_path = args.save_path + f"{args.model}_{args.dataset}/"
    with open(save_path + "info_new_" + args.dataset + ".pkl", 'wb') as f: 
        pickle.dump(info_dict, f)

if __name__ == "__main__":
    main()